import 'dart:typed_data';
import 'dart:math';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart' show rootBundle;
import 'package:image/image.dart' as imglib;
import 'package:provider/provider.dart';

import '../entities/tile.dart';
import '../provider/data.dart';
import '../utils/get_images_list.dart';

class Home extends StatelessWidget {
  static const String id = 'home';

  double _tipImageSize = 100;
  double _tileImageSize = 80;

  Future<void> resetGame(Data myProvider) async {
    myProvider.updateImage = '';
  }

  String getImageAssetName(String imageCode) {
    return 'assets/images/'+imageCode+'.png';
  }

  Future<void> getImagesList(Data myProvider) async {
    GetImagesList getImagesList;
    getImagesList = GetImagesList();
    await getImagesList.init();
    if (getImagesList.images != null) {
      myProvider.updateImages = getImagesList.images;
      shuffleImages(myProvider);
    }
  }

  Future<void> selectImage(Data myProvider, String imageCode) async {
    myProvider.updateShuffling = true;
    myProvider.updateImage = imageCode;
    myProvider.updateTiles = await splitImageInTiles(myProvider);
    myProvider.updateShuffling = false;
  }

  void shuffleImages(Data myProvider) {
    List images = myProvider.images;
    images.shuffle();
    myProvider.updateImages = images;
  }

  Container _buildImageSelectorItem(Data myProvider, String image, double size) {
    return Container(
      child: FlatButton(
        child: Container(
          decoration: BoxDecoration(
            borderRadius: BorderRadius.circular(8),
            border: Border.all(
              color: Colors.blue[200],
              width: 8,
            ),
          ),
          margin: EdgeInsets.all(2),
          child: Image(
            image: AssetImage(getImageAssetName(image)),
            width: size,
            height: size,
            fit: BoxFit.fill
          ),
        ),
        onPressed: () { selectImage(myProvider, image); },
      ),
    );
  }

  Container _buildImageSelector(Data myProvider, double size) {
    if (myProvider.images.length == 0) {
      getImagesList(myProvider);
    }
    List images = myProvider.images;

    return Container(
      padding: EdgeInsets.all(5),
      child: ListView(
        children: [
          Table(
            defaultColumnWidth: IntrinsicColumnWidth(),
            children: [
              for (var imageIndex = 0; imageIndex < images.length; imageIndex+=2)
                TableRow(
                  children: [
                    Column(
                      children: [
                        _buildImageSelectorItem(myProvider, images[imageIndex], size)
                      ]
                    ),
                    Column(
                      children: [
                        if (imageIndex+1 < images.length)
                          _buildImageSelectorItem(myProvider, images[imageIndex+1], size)
                      ]
                    ),
                  ]
                ),
            ],
          ),
        ],
      ),
    );
  }

  FlatButton _buildTilesetSizeSelectorItem(Data myProvider, int value) {
    BoxDecoration border =
      myProvider.tilesCount == value
      ? BoxDecoration(
        borderRadius: BorderRadius.circular(4),
        border: Border.all(
          color: Colors.blue[100],
          width: 4,
        ),
      )
      : null;

    return FlatButton(
      child: Container(
        padding: EdgeInsets.all(5),
        decoration: border,
        child: Text(value.toString()+'x'+value.toString()),
      ),
      onPressed: () { myProvider.updateTilesCount = value; },
    );
  }

  Container _buildTilesetSizeSelector(Data myProvider) {
    return Container(
      child: Center(
        child: Row(
          mainAxisAlignment: MainAxisAlignment.center,
          crossAxisAlignment: CrossAxisAlignment.center,
          children: [
            _buildTilesetSizeSelectorItem(myProvider, 3),
            _buildTilesetSizeSelectorItem(myProvider, 4),
          ],
        ),
      ),
    );
  }

  List<Tile> shuffleTiles(List<Tile> tiles) {
    var random = new Random();
    int tilesCount = tiles.length;

    for (int i = 0; i < 40; i++) {
      int indexTile1 = random.nextInt(tilesCount);
      int indexTile2 = random.nextInt(tilesCount);

      Tile swap = tiles[indexTile1];
      tiles[indexTile1] = tiles[indexTile2];
      tiles[indexTile2] = swap;

      int swapCol = tiles[indexTile1].currentCol;
      tiles[indexTile1].currentCol = tiles[indexTile2].currentCol;
      tiles[indexTile2].currentCol = swapCol;

      int swapRow = tiles[indexTile1].currentRow;
      tiles[indexTile1].currentRow = tiles[indexTile2].currentRow;
      tiles[indexTile2].currentRow = swapRow;
    }

    return tiles;
  }

  Future<List<Tile>> splitImageInTiles(Data myProvider) async {
    String imageAsset = getImageAssetName(myProvider.image);
    Uint8List imageData = (await rootBundle.load(imageAsset))
      .buffer
      .asUint8List();
    imglib.Image image = imglib.decodeImage(imageData);

    int x = 0, y = 0;
    int width = (image.width / myProvider.tilesCount).round();
    int height = (image.height / myProvider.tilesCount).round();

    List<Tile> tiles = List<Tile>();
    for (int i = 0; i < myProvider.tilesCount; i++) {
      for (int j = 0; j < myProvider.tilesCount; j++) {
        tiles.add(
          Tile(
            Image.memory(
              imglib.encodeJpg(
                imglib.copyCrop(image, x, y, width, height)
              )
            ),
            _tileImageSize,
            j, i,
            j, i,
          )
        );

        x += width;
      }
      x = 0;
      y += height;
    }

    return shuffleTiles(tiles);
  }

  bool _checkTilesetIsOrdered(List<Tile> tiles) {
    for (Tile tile in tiles) {
      if (
        (tile.currentRow != tile.originalRow)
        ||
        (tile.currentCol != tile.originalCol)
      ) {
        return false;
      }
    }

    return true;
  }

  Container _buildTilesetWidget(Data myProvider) {
    List tiles = myProvider.tiles;

    Color borderColor = _checkTilesetIsOrdered(tiles) ? Colors.green : Colors.orange;
    int tileIndex = 0;

    Table tileset = Table(
      defaultColumnWidth: IntrinsicColumnWidth(),

      border: TableBorder.all(
        color: Colors.black,
        style: BorderStyle.solid,
        width: 2,
      ),
      children: [
        for (var row = 0; row < myProvider.tilesCount; row++)
          TableRow(children: [
            for (var col = 0; col < myProvider.tilesCount; col++)
              Column(children: [tiles[tileIndex++].widget(myProvider)]),
          ]),
      ]
    );

    return Container(
      margin: EdgeInsets.all(8),
      padding: EdgeInsets.all(8),
      decoration: BoxDecoration(
        color: borderColor,
        borderRadius: BorderRadius.circular(8),
        border: Border.all(
          color: borderColor,
          width: 8,
        ),
      ),
      child: tileset,
    );
  }

  Container _buildTipWidget(Data myProvider) {
    return Container(
      margin: EdgeInsets.all(4),
      padding: EdgeInsets.all(4),
      decoration: BoxDecoration(
        color: Colors.blue,
        borderRadius: BorderRadius.circular(4),
        border: Border.all(
          color: Colors.blue,
          width: 4,
        ),
      ),
      child: Image(
        image: AssetImage(getImageAssetName(myProvider.image)),
        width: _tipImageSize,
        height: _tipImageSize,
        fit: BoxFit.fill
      ),
    );
  }

  Container _buildGameWidget(Data myProvider) {
    return Container(
      child: Column(
        mainAxisSize: MainAxisSize.min,
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          _buildTilesetWidget(myProvider),
          SizedBox(height: 20),
          _buildTipWidget(myProvider),
        ],
      ),
    );
  }

  Container _buildShufflingIndicatorWidget() {
    return Container(
      child: Column(
        mainAxisSize: MainAxisSize.min,
        mainAxisAlignment: MainAxisAlignment.center,
        children: [
          Text(
            '⏳',
            style: TextStyle(
              fontSize: 60,
              fontWeight: FontWeight.w600,
              color: Colors.black,
            ),
          ),
        ],
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    Data _myProvider = Provider.of<Data>(context);

    var content;

    double _selectorImageSize = MediaQuery.of(context).size.width / 3;

    if (_myProvider.shuffling) {
      content = _buildShufflingIndicatorWidget();
    } else {
      if (_myProvider.image == '') {
        content = _buildImageSelector(_myProvider, _selectorImageSize);
      } else {
        content = _buildGameWidget(_myProvider);
      }
    }

    List<Widget> menuActions = [
      IconButton(
        icon: const Icon(Icons.shuffle),
        onPressed: () => shuffleImages(_myProvider),
      ),
      _buildTilesetSizeSelector(_myProvider),
    ];

    List<Widget> gameActions = [
      IconButton(
        icon: const Icon(Icons.arrow_left_outlined),
        onPressed: () => resetGame(_myProvider),
      )
    ];

    return Scaffold(
      appBar: AppBar(
        title: Text('Puzzle'),
        actions: _myProvider.image == '' ? menuActions : gameActions,
      ),
      body: SafeArea(
        child: Center(
          child: content
        ),
      )
    );
  }
}
